/* wptClipDecryptDlg.cpp - Clipboard decrypt dialog
 *	Copyright (C) 2000-2005 Timo Schulz
 *	Copyright (C) 2005 g10 Code GmbH
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>

#include "wptTypes.h"
#include "wptW32API.h"
#include "wptAgent.h"
#include "wptNLS.h"
#include "wptGPG.h"
#include "wptVersion.h"
#include "wptErrors.h"
#include "wptCommonCtl.h"
#include "wptContext.h"
#include "wptDlgs.h"
#include "wptKeylist.h"
#include "wptFileManager.h"
#include "wptUTF8.h"
#include "resource.h"

bool secret_key_available (gpgme_recipient_t rset);

/* Return the primary user-ID of the key with the keyid @keyid. 
   Caller must free string. */
char*
get_key_userid (const char *keyid)
{
    gpgme_key_t key;
    char *p, *uid;

    if (get_pubkey (keyid, &key))
	return m_strdup (_("user ID not found"));
    uid = utf8_to_native (key->uids->uid);
    if (!uid)
	uid = strdup (_("user ID not found"));
    p = new char[strlen (uid) + 4 + 8];
    if (!p)
	BUG (NULL);
    sprintf (p, "\n    \"%s\"", uid);
    safe_free (uid);
    return p;
}


/* Decrypt the clipboard contents and on success
   replace the data with the plaintext.
   Return value: 0 on success. */
gpgme_error_t
gpgme_op_clip_decrypt (gpgme_ctx_t ctx)
{
    gpgme_error_t err;
    gpgme_data_t ciph = NULL;
    gpgme_data_t plain = NULL;
    
    err = gpg_data_new_from_clipboard (&ciph, 0);
    if (err)
	return err;

    err = gpgme_data_new (&plain);
    if (err) {
	gpgme_data_release (ciph);
	return err;
    }

    err = gpgme_op_decrypt_verify (ctx, ciph, plain);

    gpg_data_release_and_set_clipboard (plain, 0);
    gpgme_data_release (ciph);
    return err;
}


/* Convenient function to provide clipboard decryption.
   @hwnd is the parent window used for showing messsages.
   Return value: 0 on success. */
gpgme_error_t
clip_decrypt_dlg (HWND hwnd)
{
    gpgme_error_t err;
    gpgme_ctx_t ctx = NULL;
    gpgme_signature_t sig = NULL;
    gpgme_decrypt_result_t res;
    gpgme_verify_result_t sigres;
    passphrase_cb_s pwd;
    const char *s;
    char *uid;
    int pgp_type = 0;
    int novalid = 0;

    /* allow to verify data generated by 'gpg -a --sign foo' */
    if (fm_assume_onepass_sig (NULL) == 0) {
	/* XXX: addtitional steps needed? */
    }

    err = gpgme_new (&ctx);
    if (err)
	BUG (NULL);
    set_gpg_passphrase_cb (&pwd, ctx, GPG_CMD_DECRYPT, hwnd, _("Decryption"));
    gpg_get_recipients (NULL, &pwd.recipients);

    err = gpgme_op_clip_decrypt (ctx);
    if (pwd.cancel)
	goto leave;
    if (gpgme_err_code (err) ==  GPG_ERR_BAD_PASSPHRASE)
	agent_del_cache (pwd.keyid);

    res = gpgme_op_decrypt_result (ctx);
    if (err && res->recipients && !secret_key_available (res->recipients)) {
	gpgme_recipient_t r = res->recipients;
	uid = get_key_userid (r->keyid+8);
	log_box (_("Decryption"), MB_ERR, 
		 _("Encrypted with %s key, ID %s.%s\n"
		   "Decryption failed: secret key not available."), 
		   get_key_pubalgo (r->pubkey_algo), r->keyid+8, uid);
	free_if_alloc (uid);
	goto leave;
    }
    else if (res->unsupported_algorithm) {
	log_box (_("Decryption"), MB_ERR, _("Unsupported algorithm: %s"), 
		 res->unsupported_algorithm);
    }
    else if (err) {
	gpg_clip_get_pgptype (&pgp_type);
	if (gpgme_err_code (err) == GPG_ERR_NO_DATA && (pgp_type & PGP_MESSAGE))
	    msg_box (hwnd, _("Broken OpenPGP message (maybe: quoted printable "
			     "character in armor)."), _("Decryption"), MB_INFO);
	else
	    msg_box (hwnd, gpgme_strerror (err), _("Decryption"), MB_ERR);
	goto leave;
    }

    if (0) { /* XXX: Bad MDC */
	s = _("WARNING: encrypted message has been manipulated!\n"
	    "\n"
	    "Do *NOT* trust any text or data output from this file!\n"
	    "It is likely, the data was corrupted during the transport\n"
	    "but it might be also possible that this is part of an attack.");
	msg_box (hwnd, s, _("*** IMPORTANT ***"), MB_INFO);
    }

    show_msg (GetDesktopWindow (), 1500, _("GnuPG Status: Finished"));

    sigres = gpgme_op_verify_result (ctx);
    if (sigres && sigres->signatures) {
	gpgme_key_t key=NULL;
	const char *keyid;

	sig = sigres->signatures;
	if (!sig->fpr)
	    BUG (NULL);
	if (strlen (sig->fpr) > 16)
	    keyid = strlen (sig->fpr) == 40? sig->fpr+24 : sig->fpr+16;
	else
	    keyid = sig->fpr;

	get_pubkey (keyid, &key);
	if (key) {
	    if (key->owner_trust == GPGME_VALIDITY_FULL || 
		key->owner_trust == GPGME_VALIDITY_ULTIMATE)
		s = _("Signature Status: Created with a fully trusted key");
	    else if (key->owner_trust == GPGME_VALIDITY_MARGINAL)
		s = _("Signature Status: Created with a marginal trusted key");
	    else if (key->owner_trust == GPGME_VALIDITY_NEVER) {
		novalid = 1;
		s =  _("Signature Status: Created with an UNTRUSTED key");
	    }
	    else
		s = _("Signature Status: Created with an undefined trusted key");
	    uid = utf8_to_native (key->uids->uid);
	}
	else {
	    s = "";
	    uid = strdup (_("user ID not found"));
	}
	log_box (_("WinPT Verify"), MB_OK, 
		 _("%s\n"
		   "%s\n"
                   "Signature made: %s\n"
                   "From \"%s\" using key ID 0x%s"
		   "%s %s"),
		   s, get_gpg_sigstat (sig->summary), 
		   strtimestamp (sig->timestamp),
		   uid, keyid+8, 
		   novalid? "\nPrimary key fingerprint: " : "",
		   novalid? get_key_fpr (key) : "");
	safe_free (uid);
    }
    
leave:
    release_gpg_passphrase_cb (&pwd);
    gpgme_release (ctx);
    return err;
}
