/* wptPassphraseDlg.cpp - Dialog to get the passphrase
 *	Copyright (C) 2001-2005 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>

#include "resource.h"
#include "wptGPG.h"
#include "wptCommonCtl.h"
#include "wptContext.h"
#include "wptDlgs.h"
#include "wptNLS.h"
#include "wptW32API.h"
#include "wptUTF8.h"
#include "wptVersion.h"
#include "wptTypes.h"


struct passphrase_s {
    char *title;    /* title of the dialog. */
    char pwd[256];  /* the actual passphrase. */
    int strict;	    /* do a simple check how good the passphrase is. */
    int repeat;	    /* Indicate last try was wrong. */   
    int cancel;	    /* 1 if user cancelled operation. */
    int not_empty;
};


/* Dialog procedure to request a passphrase from the user. */
static BOOL CALLBACK
passwd_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static passphrase_s *pwd;
    int nbytes;
    
    switch (msg) {
    case WM_INITDIALOG:
        pwd = (passphrase_s *)lparam;
        if (pwd == NULL)
            dlg_fatal_error (dlg, "Could not get dialog param!");
        SetWindowText (dlg, _("Passphrase Dialog"));
	CheckDlgButton (dlg, IDC_PASSWD_HIDE, BST_CHECKED);
        if (pwd->title)
            SetWindowText (dlg, pwd->title);
	if (pwd->repeat)
	    SetDlgItemText (dlg, IDC_PASSWD_INFO, _("Repeat Passphrase"));
	else
	    SetDlgItemText (dlg, IDC_PASSWD_INFO, _("Enter Passphrase"));
	SetDlgItemText (dlg, IDC_PASSWD_HIDE, _("&Hide Typing"));
	SetDlgItemText (dlg, IDCANCEL, _("&Cancel"));
        SetFocus (GetDlgItem (dlg, IDC_PASSWD_PWD ));
	center_window2 (dlg, NULL, HWND_TOPMOST);
	center_window (dlg, NULL);
        SetForegroundWindow (dlg);
        return FALSE;

    case WM_ACTIVATE:
	safe_edit_control_init (dlg, IDC_PASSWD_PWD);
	break;

    case WM_DESTROY:
	safe_edit_control_free (dlg, IDC_PASSWD_PWD);
	break;
        
    case WM_COMMAND:
	if (HIWORD (wparam) == BN_CLICKED && 
	    LOWORD (wparam) == IDC_PASSWD_HIDE) {
	    HWND hwnd = GetDlgItem (dlg, IDC_PASSWD_PWD);
	    int hide = IsDlgButtonChecked (dlg, IDC_PASSWD_HIDE);
	    SendMessage (hwnd, EM_SETPASSWORDCHAR, hide? '*' : 0, 0);
	    SetFocus (hwnd);
	}

        switch (LOWORD (wparam)) {
        case IDOK:
            pwd->cancel = 0;
	    nbytes = SafeGetDlgItemText (dlg, IDC_PASSWD_PWD, 
					 pwd->pwd, DIM (pwd->pwd)-1);
	    if (!nbytes && pwd->not_empty) {
		msg_box (dlg, _("Please enter a passphrase."), _("Passphrase Dialog"), MB_ERR);
		return TRUE;
	    }

            if (!nbytes)
                strcpy (pwd->pwd, "");
	    else if (pwd->strict && check_passwd_quality (pwd->pwd, 0)) {
		int id = msg_box (dlg, _("Your passphrase should be at least 8 characters"
					 " long\nand should contain non-alphabetic characters."
					 "\n\nStill proceed?"),
				  _("Key Generation"), MB_ICONWARNING|MB_YESNO);
		if (id == IDNO)
		    return TRUE;
	    }
            SetDlgItemText (dlg, IDC_PASSWD_PWD, "");
            EndDialog (dlg, TRUE);
            return TRUE;
            
        case IDCANCEL:
            pwd->cancel = 1;
            EndDialog (dlg, FALSE);
            return TRUE;
        }
        break;
    }
    
    return FALSE;
}


/* Request a passphrase from the user. @title is the title of the
   dialog and @ret_cancel is true if user cancelled the operation.
   Return value: the passphrase or NUL for an error. */
char*
request_passphrase (const char *title, int flags, int *ret_cancel)
{
    passphrase_s pass;
    char *p;

    *ret_cancel = 0; /* reset */
    memset (&pass, 0, sizeof (pass));
    if (title && *title) {
	pass.title = m_strdup (title);
	if (!pass.title)
	    BUG (0);
    }
    pass.repeat = flags & PASSDLG_INIT? 0 : 1;
    pass.strict = flags & PASSDLG_STRICT? 1 : 0;
    pass.not_empty = flags & PASSDLG_NOTEMPTY? 1: 0;
    DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_PASSWD, glob_hwnd,
		    passwd_dlg_proc, (LPARAM)&pass);
    if (pass.cancel == 1) {
	*ret_cancel = 1;
	return NULL;
    }
    p = m_strdup (pass.pwd);
    if (!p)
	BUG (0);
    free_if_alloc (pass.title);
    memset (&pass, 0, sizeof (pass));
    return p;
}


/* Request a passphrase from the user but also confirm the passphrase
   to make sure there is no typo. Arguments same as in the normal version
   of the function. */
char*
request_passphrase2 (const char *title, int flags, int *ret_cancel)
{
    char *pass1, *pass2;
    int equal = 0, cancel = 0;

    *ret_cancel = 1;
    while (!equal) {
	pass1 = request_passphrase (title, PASSDLG_INIT|flags, &cancel);
	if (cancel)
	    return NULL;
	pass2 = request_passphrase (title, PASSDLG_REPEAT, &cancel);
	if (cancel) {
	    sfree_if_alloc (pass1);
	    return NULL;
	}

	if (strcmp (pass1, pass2)) {
	    msg_box (NULL, _("Passphrases do not match. Please try again."),
		     title, MB_INFO);
	    sfree_if_alloc (pass1);
	    sfree_if_alloc (pass2);
	}
	else {
	    if (is_8bit_string (pass1)) {
		msg_box (NULL, _("The passphrase contains 8-bit characters.\n"
				 "It is not suggested to use charset specific characters."),
			 title, MB_ERR);
		equal = 0;
		sfree_if_alloc (pass1);
		sfree_if_alloc (pass2);
	    }
	    else
		equal = 1;
	}
    }
    sfree_if_alloc (pass2);
    *ret_cancel = 0;
    return pass1;
}
